@@ -773,15 +773,19 @@ impl<R: AsyncRead + Unpin> BinFuture<R> {
         self.len == 0
     }
 
-    /// Panics if buf length is less than the value returned by `len()`
-    pub async fn read_all(mut self, buf: &mut [u8]) -> IoResult<R> {
-        self.reader
-            .read_exact(&mut buf[..self.len])
-            .await
-            .map(|_| self.reader)
+    /// Panics if buf length is greater than bytes remaining
+    pub async fn read_exact(&mut self, buf: &mut [u8]) -> IoResult<()> {
+        assert!(buf.len() <= self.len);
+        self.reader.read_exact(buf).await.map(|_| {
+            self.len -= buf.len();
+        })
     }
 
-    pub fn into_inner(self) -> R {
+    pub async fn read_all(mut self, buf: &mut [u8]) -> IoResult<R> {
+        self.read_exact(buf).await.map(|_| self.end())
+    }
+
+    pub fn end(self) -> R {
         assert_eq!(self.len, 0);
         self.reader
     }
@@ -824,6 +828,10 @@ impl<R: AsyncRead + Unpin> StringFuture<R> {
         })
     }
 
+    pub async fn read_exact(&mut self, buf: &mut [u8]) -> IoResult<()> {
+        self.0.read_exact(buf).await
+    }
+
     pub async fn read_all(self, buf: &mut [u8]) -> IoResult<R> {
         self.0.read_all(buf).await
     }
@@ -880,6 +888,10 @@ impl<R: AsyncRead + Unpin> ExtFuture<R> {
         self.bin.into_vec().await.map(|(v, r)| (ty, v, r))
     }
 
+    pub async fn read_exact(&mut self, buf: &mut [u8]) -> IoResult<()> {
+        self.bin.read_exact(buf).await
+    }
+
     pub async fn read_all(self, buf: &mut [u8]) -> IoResult<R> {
         self.bin.read_all(buf).await
     }
